#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <cooperative_groups.h>
#include <cooperative_groups/reduce.h>

namespace cg = cooperative_groups;

template<int STRIDE>
__device__ __forceinline__ float warp_reduce_sum(float val) {
    #pragma unroll
    for (int i = STRIDE; i > 0; i >>= 1) {
        val += __shfl_down_sync(0xffffffff, val, i);
    }
    return val;
}

__device__ __forceinline__ void linear_transformation(
    const float* __restrict__ a,
    const float* __restrict__ w,
    int row,
    int col,
    int K,
    float& sum
) {
    const int tid = threadIdx.x % 32;
    const float* a_row = a + row * K;
    const float* w_row = w + col * K;

    #pragma unroll 4
    for (int k = tid; k < K; k += 32) {
        sum += a_row[k] * w_row[k];
    }
}

__global__ void linear_tanh_kernel(
    const float* __restrict__ A,
    const float* __restrict__ weight,
    const float* __restrict__ bias,
    float* __restrict__ out,
    int B, int K, int M
) {
    const int global_tid = blockIdx.x * blockDim.x + threadIdx.x;
    const int warp_id = global_tid / 32;
    const int row = warp_id / M;
    const int col = warp_id % M;

    if (row >= B) return;

    float sum = 0.0f;
    linear_transformation(A, weight, row, col, K, sum);
    sum = warp_reduce_sum<16>(sum);

    if (threadIdx.x % 32 == 0) {
        out[row * M + col] = tanhf(sum + bias[col]);
    }
}

__global__ void concat_kernel(
    const float* __restrict__ x,
    const float* __restrict__ hidden,
    float* __restrict__ combined,
    int batch_size,
    int x_size,
    int hidden_size,
    int total_elements
) {
    const int idx = blockIdx.x * blockDim.x + threadIdx.x;
    const int combined_width = x_size + hidden_size;
    
    for (int i = idx; i < total_elements; i += blockDim.x * gridDim.x) {
        const int row = i / combined_width;
        const int col = i % combined_width;
        combined[i] = (col < x_size) 
            ? x[row * x_size + col] 
            : hidden[row * hidden_size + (col - x_size)];
    }
}

torch::Tensor module_fn_cuda(
    torch::Tensor x,
    torch::Tensor i2h_weight,
    torch::Tensor i2h_bias,
    torch::Tensor h2o_weight,
    torch::Tensor h2o_bias,
    torch::Tensor hidden
) {
    x = x.contiguous().cuda();
    i2h_weight = i2h_weight.contiguous().cuda();
    i2h_bias = i2h_bias.contiguous().cuda();
    h2o_weight = h2o_weight.contiguous().cuda();
    h2o_bias = h2o_bias.contiguous().cuda();
    hidden = hidden.contiguous().cuda();

    const int batch_size = x.size(0);
    const int x_size = x.size(1);
    const int hidden_size = hidden.size(1);
    const int combined_size = x_size + hidden_size;

    auto combined = torch::empty({batch_size, combined_size}, x.options());
    
    // Concatenation kernel
    const int concat_blocks = (batch_size * combined_size + 255) / 256;
    concat_kernel<<<concat_blocks, 256>>>(
        x.data_ptr<float>(),
        hidden.data_ptr<float>(),
        combined.data_ptr<float>(),
        batch_size,
        x_size,
        hidden_size,
        batch_size * combined_size
    );

    // Custom linear kernel
    const int M = i2h_weight.size(0);
    const int num_warps = batch_size * M;
    const int threads_per_block = 256;
    const int blocks = (num_warps * 32 + threads_per_block - 1) / threads_per_block;
    
    auto hidden_new = torch::empty({batch_size, M}, x.options());
    linear_tanh_kernel<<<blocks, threads_per_block>>>(
        combined.data_ptr<float>(),
        i2h_weight.data_ptr<float>(),
        i2h_bias.data_ptr<float>(),
        hidden_new.data_ptr<float>(),
        batch_size,
        combined_size,
        M
    );

    // Output layer (keep as optimized torch::addmm)
    return torch::addmm(h2o_bias, hidden_new, h2o_weight.t());
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &module_fn_cuda, "Optimized Modular RNN Forward (CUDA)");
}